//! [Monad] for passing and combining metadata related to the evaluation of values.
//!
//! For "either"/"sum" wrapping, see [`result`].
//!
//! For no extra data, see [`solo`].
//!
//! [`result`]: super::result
//! [`solo`]: super::solo

use crate::func::class_prelude::*;

/// Metadata type.
pub trait Effect: Send {
    /// Used in [`Pure::pure`].
    fn e_pure() -> Self;

    /// Used in [`ApplicativeLA2::la2`] and other [`Applicative`] methods.
    fn e_seq(el: Self, er: Self) -> Self;

    /// Used in [`Monad::bind`] and other [`Monad`] methods.
    fn e_after(self, effect: Self) -> Self;
}

/// Value and related metadata.
#[derive(Clone)]
pub struct WithEffect<A, E> {
    pub value: A,
    pub effect: E,
}

impl<A, E: Effect> WithEffect<A, E> {
    fn e_after(self, effect: E) -> Self {
        WithEffect {
            value: self.value,
            effect: self.effect.e_after(effect),
        }
    }
}

pub struct EffectInstance<E: Send>(E);

impl<E: Send + Sync + Clone> SharedFunctorAny for EffectInstance<E> {
    type SharedAny<'a, A: 'a + Send + Sync + Clone> = Self::FAny<'a, A>
    where
        Self: 'a;

    fn share<'a, A: 'a + Send + Sync + Clone>(fa: Self::FAny<'a, A>) -> Self::SharedAny<'a, A>
    where
        Self: 'a,
    {
        fa
    }

    fn unshare<'a, A: 'a + Send + Sync + Clone>(sa: Self::SharedAny<'a, A>) -> Self::FAny<'a, A>
    where
        Self: 'a,
    {
        sa
    }
}

impl<E: Send> WeakFunctorAny for EffectInstance<E> {
    type FAny<'a, A: 'a + Send> = WithEffect<A, E>
    where
        Self: 'a;
}

impl<'a, E: 'a + Send> Functor<'a> for EffectInstance<E> {
    fn fmap<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> B) -> Self::F<B> {
        WithEffect {
            value: f(fa.value),
            effect: fa.effect,
        }
    }

    fn replace<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, b: B) -> Self::F<B> {
        drop(fa.value);
        WithEffect {
            value: b,
            effect: fa.effect,
        }
    }
}

impl<'a, E: 'a + Effect> Pure<'a> for EffectInstance<E> {
    fn pure<A: 'a + Send>(a: A) -> Self::F<A> {
        WithEffect {
            value: a,
            effect: E::e_pure(),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeSeq<'a> for EffectInstance<E> {
    fn seq<A: 'a + Send, B: 'a + Send>(
        ff: Self::F<impl 'a + Send + FnOnce(A) -> B>,
        fa: Self::F<A>,
    ) -> Self::F<B> {
        WithEffect {
            value: (ff.value)(fa.value),
            effect: E::e_seq(ff.effect, fa.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeLA2<'a> for EffectInstance<E> {
    fn la2<A: 'a + Send, B: 'a + Send, C: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + Send + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        WithEffect {
            value: f(fa.value, fb.value),
            effect: E::e_seq(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeTuple<'a> for EffectInstance<E> {
    fn tuple<A: 'a + Send, B: 'a + Send>((fa, fb): (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        WithEffect {
            value: (fa.value, fb.value),
            effect: E::e_seq(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeSelect<'a> for EffectInstance<E> {}

impl<'a, E: 'a + Effect> Applicative<'a> for EffectInstance<E> {
    fn discard_first<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        drop(fa.value);
        WithEffect {
            value: fb.value,
            effect: E::e_seq(fa.effect, fb.effect),
        }
    }

    fn discard_second<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        drop(fb.value);
        WithEffect {
            value: fa.value,
            effect: E::e_seq(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> Monad<'a> for EffectInstance<E> {
    fn bind<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + FnOnce(A) -> Self::F<B>,
    ) -> Self::F<B> {
        f(fa.value).e_after(fa.effect)
    }

    fn iterate<B: 'a + Send>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        let mut effect = E::e_pure();
        loop {
            let fa = f.next();
            effect = fa.effect.e_after(effect);
            match fa.value {
                ControlFlow::Continue(next_f) => f = next_f,
                ControlFlow::Break(b) => return WithEffect { value: b, effect },
            }
        }
    }

    fn join<A: 'a + Send>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        ffa.value.e_after(ffa.effect)
    }
}

impl<'a, E: 'a + Send> LocalFunctor<'a> for EffectInstance<E> {
    fn stuff<A: 'a + Send, T: Pure<'a>>(fa: Self::F<T::F<A>>) -> T::F<Self::F<A>> {
        T::fmap(fa.value, |a| WithEffect {
            value: a,
            effect: fa.effect,
        })
    }
}
